package java.util.concurrent;

import java.util.Spliterator;
import java.util.Map;
import java.util.Set;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.Collection;
import java.util.Comparator;
import sun.misc.Unsafe;
import java.io.Serializable;
import java.util.NavigableSet;
import java.util.AbstractSet;

public class ConcurrentSkipListSet<E> extends AbstractSet<E> implements NavigableSet<E>, Cloneable, Serializable
{
    private static final long serialVersionUID = -2479143111061671589L;
    private final ConcurrentNavigableMap<E, Object> m;
    private static final Unsafe UNSAFE;
    private static final long mapOffset;
    
    public ConcurrentSkipListSet() {
        this.m = new ConcurrentSkipListMap();
    }
    
    public ConcurrentSkipListSet(Comparator<? super E> comparator) {
        this.m = new ConcurrentSkipListMap(comparator);
    }
    
    public ConcurrentSkipListSet(Collection<? extends E> c) {
        this.m = new ConcurrentSkipListMap();
        this.addAll(c);
    }
    
    public ConcurrentSkipListSet(SortedSet<E> c) {
        this.m = new ConcurrentSkipListMap(c.comparator());
        this.addAll(c);
    }
    
    ConcurrentSkipListSet(ConcurrentNavigableMap<E, Object> m) {
        this.m = m;
    }
    
    public ConcurrentSkipListSet<E> clone() {
        try {
            ConcurrentSkipListSet<Object> set = (ConcurrentSkipListSet)super.clone();
            set.setMap(new ConcurrentSkipListMap(this.m));
            return set;
        }
        catch (CloneNotSupportedException ex) {
            throw new InternalError();
        }
    }
    
    @Override
    public int size() {
        return this.m.size();
    }
    
    @Override
    public boolean isEmpty() {
        return this.m.isEmpty();
    }
    
    @Override
    public boolean contains(Object o) {
        return this.m.containsKey(o);
    }
    
    @Override
    public boolean add(E e) {
        return this.m.putIfAbsent(e, Boolean.TRUE) == null;
    }
    
    @Override
    public boolean remove(Object o) {
        return this.m.remove(o, Boolean.TRUE);
    }
    
    @Override
    public void clear() {
        this.m.clear();
    }
    
    @Override
    public Iterator<E> iterator() {
        return this.m.navigableKeySet().iterator();
    }
    
    @Override
    public Iterator<E> descendingIterator() {
        return this.m.descendingKeySet().iterator();
    }
    
    @Override
    public boolean equals(Object o) {
        if (o == this) {
            return true;
        }
        if (!(o instanceof Set)) {
            return false;
        }
        Collection<Object> c = (Collection)o;
        try {
            return this.containsAll(c) && c.containsAll(this);
        }
        catch (ClassCastException ex) {
            return false;
        }
        catch (NullPointerException ex2) {
            return false;
        }
    }
    
    @Override
    public boolean removeAll(Collection<?> collection) {
        boolean n = false;
        Iterator<Object> iterator = collection.iterator();
        while (iterator.hasNext()) {
            if (this.remove(iterator.next())) {
                n = true;
            }
            else {
                continue;
            }
        }
        return n;
    }
    
    @Override
    public E lower(E e) {
        return this.m.lowerKey(e);
    }
    
    @Override
    public E floor(E e) {
        return this.m.floorKey(e);
    }
    
    @Override
    public E ceiling(E e) {
        return this.m.ceilingKey(e);
    }
    
    @Override
    public E higher(E e) {
        return this.m.higherKey(e);
    }
    
    @Override
    public E pollFirst() {
        Map.Entry<Object, Object> pollFirstEntry = this.m.pollFirstEntry();
        return (pollFirstEntry == null) ? null : pollFirstEntry.getKey();
    }
    
    @Override
    public E pollLast() {
        Map.Entry<Object, Object> pollLastEntry = this.m.pollLastEntry();
        return (pollLastEntry == null) ? null : pollLastEntry.getKey();
    }
    
    @Override
    public Comparator<? super E> comparator() {
        return this.m.comparator();
    }
    
    @Override
    public E first() {
        return this.m.firstKey();
    }
    
    @Override
    public E last() {
        return this.m.lastKey();
    }
    
    @Override
    public NavigableSet<E> subSet(E e, boolean b, E e2, boolean b2) {
        return new ConcurrentSkipListSet(this.m.subMap(e, b, e2, b2));
    }
    
    @Override
    public NavigableSet<E> headSet(E e, boolean b) {
        return new ConcurrentSkipListSet(this.m.headMap(e, b));
    }
    
    @Override
    public NavigableSet<E> tailSet(E e, boolean b) {
        return new ConcurrentSkipListSet(this.m.tailMap(e, b));
    }
    
    @Override
    public NavigableSet<E> subSet(E e, E e2) {
        return this.subSet(e, true, e2, false);
    }
    
    @Override
    public NavigableSet<E> headSet(E e) {
        return this.headSet(e, false);
    }
    
    @Override
    public NavigableSet<E> tailSet(E e) {
        return this.tailSet(e, true);
    }
    
    @Override
    public NavigableSet<E> descendingSet() {
        return new ConcurrentSkipListSet(this.m.descendingMap());
    }
    
    @Override
    public Spliterator<E> spliterator() {
        if ((this.m instanceof ConcurrentSkipListMap)) {
            return ((ConcurrentSkipListMap)this.m).keySpliterator();
        }
        return (Spliterator)((ConcurrentSkipListMap.SubMap)this.m).keyIterator();
    }
    
    private void setMap(ConcurrentNavigableMap<E, Object> x) {
        ConcurrentSkipListSet.UNSAFE.putObjectVolatile(this, ConcurrentSkipListSet.mapOffset, x);
    }
    
    static {
        try {
            UNSAFE = Unsafe.getUnsafe();
            mapOffset = ConcurrentSkipListSet.UNSAFE.objectFieldOffset(ConcurrentSkipListSet.class.getDeclaredField("m"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
}
